# coding=utf-8

import logging
module_logger = logging.getLogger("pyrst.client")
module_logger.setLevel(logging.DEBUG)


class PyrstException(Exception):
    """
    Generic, abstract class of exception.
    """

    def __init__(self):
        self.logger = logging.getLogger("pyrst.client")
        self.logger.warning(self.__repr__())


class TokenException(PyrstException):
    """
    Raised when there is no token saved in the instance and a function that
    requires a token (i.e. any function other than login()) is called."
    """

    def __repr__(self):
        return "Cannot perform this operation without authentication token. Use" \
               " the login() method to obtain one."


class AuthException(PyrstException):
    """
    Raised when the user is not authorised.
    """

    def __repr__(self):
        return "Not authorised."


class ConnectionException(PyrstException):
    """
    Raised when the client could not connect for a network error.
    """

    def __repr__(self):
        return "Connection error."


class SpaceIDException(PyrstException):
    """
    Raised where a space ID is provided that does not meet the formal criteria
    for a space ID (36 characters separated by hyphens).
    """

    def __repr__(self):
        return "You have provided an incorrect space ID. A valid Birst space ID" \
               "is 36 characters long and consists of five groups of hexadecimal" \
               "characters separated by hyphens."


class MissingCredentialsException(PyrstException):
    """
    Raised where an operation that requires credentials (e.g. login()) is
    called without providing the appropriate credentials, either directly
    or via a configuration file.
    """

    def __repr__(self):
        return "You need to provide a password and a username, either via your" \
               " configuration file or at the time of creating the Birst client " \
               "object."


# coding=utf-8
#! usr/bin/env/python

import argparse
import json
import yaml
from base64 import b64decode

from pyrst.client import BirstClient
from pyrst.exceptions import MissingCredentialsException
from pyrst.handlers import CsvHandler, JsonHandler, DfHandler


parser = argparse.ArgumentParser(description='A Birst client.')

parser.add_argument('-q', '--query',
                    required=True)
parser.add_argument('-s', '--space',
                    required=True,
                    type=str)
parser.add_argument('-u', '--username',
                    required=False,
                    type=str)
parser.add_argument('-p', '--password',
                    required=False,
                    type=str)
parser.add_argument('-i', '--instance',
                    required=False,
                    type=str)
parser.add_argument('-f', '--configfile',
                    required=False,
                    type=str)
parser.add_argument('-o', '--outputfile',
                    required=False,
                    type=str)
parser.add_argument('-H', '--handler',
                    required=False,
                    default="JSON",
                    choices=[None, "CSV", "JSON", "DF", "XLS"])


def main():
    args = parser.parse_args()

    if args.username and args.password:
        cl = BirstClient(user=args.username,
                         password=args.password,
                         instance=args.instance if args.instance else "app2102")
    elif args.configfile:
        with open(args.configfile) as config_file:
            config = yaml.load(config_file)
            password = b64decode(config["password"]) if config["password_is_encrypted"] else config["password"]

        cl = BirstClient(user=config["username"],
                         password=password,
                         instance=getattr(config, "instance", "app2102"))
    else:
        raise MissingCredentialsException

    cl.login()

    _handler_map = {"CSV": CsvHandler,
                    "JSON": JsonHandler,
                    "DF": DfHandler,
                    "XLS": DfHandler}

    _res = cl.retrieve(space=args.space,
                       query=args.query,
                       handler=_handler_map[args.handler])

    if args.outputfile:
        if args.handler == "XLS":
            _res.to_excel(args.outputfile,
                          index=False,
                          header=True,
                          na_rep="N/A",
                          float_format="%.2f")
        else:
            with open(args.outputfile, 'w') as output_file:
                if args.handler == "JSON":
                    json.dump(_res, output_file)
                elif args.handler == "DF":
                    output_file.write(_res.to_string())
                elif args.handler == "XLS":
                    _res.to_excel()
                else:
                    output_file.write(str(_res))
    else:
        print _res

if __name__ == '__main__':
    main()


# coding=utf-8
from exceptions import TokenException


def check_token(func):
    """
    Checks whether the instance variable on the object from which the function
    `func` is called has its internals set.

    :param func: function
    :return: function
    :raise: TokenException if no token is saved in the instance.
    """

    def inner(self, *args, **kwargs):
        if self.token is None:
            raise TokenException
        else:
            return func(self, *args, **kwargs)
    return inner



# coding=utf-8

import pandas as pd
import logging
import json

pd.set_option('display.float_format', lambda x: '%.3f' % x)

module_logger = logging.getLogger("pyrst.client")
module_logger.setLevel(logging.DEBUG)


class Handler(object):
    """
    Abstract class that represents Handler objects.

    In Pyrst, a Handler object receives the raw query result and converts it
    into a specified format, such as a pandas DataFrame or a D3 compatible
    JSON file.

    A Handler needs to have a process() function that takes the query_output
    as its argument. The return of that function is going to be the output of
    the Handler.
    """

    def process(self,
                query_output):
        """
        Default query output processor, returns the query output in its raw
        form.

        :param query_output: raw query output
        :return: raw query output instance
        :rtype: str
        """
        return query_output


class DfHandler(Handler):
    """
    Handler that returns a `pandas` `DataFrame`.
    """
    def __init__(self):
        self.logger = logging.getLogger("pyrst.client")
        self.logger.info("Setting up handler...")

    def process(self,
                query_output):
        """
        Query output processor that returns a `pandas` `DataFrame` object.

        :param query_output: raw query output
        :return: `pandas` `DataFrame` object representing the result
        :rtype: DataFrame
        """
        self.logger.debug("Processing query output...")

        _series = [each[0] for each in query_output["rows"]]

        _typemap = {12: "object", 8: "float"}

        _types = {}
        for k, v in enumerate(query_output["dataTypes"]):
            _types[k] = _typemap[v]

        _df = pd.DataFrame(_series)

        _df.columns = query_output["columnNames"]

        for k, v in enumerate(_df.columns):
            _df[v] = _df[v].astype(_types[k])

        self.logger.debug("Processing columns {}.".format(', '.join(list(_df.columns))))
        return _df


class JsonHandler(Handler):
    """
    Handler that returns a JSON file, ready to be ingested by D3.

    Currently, this is implemented 'lazily', i.e. the DfHandler is invoked,
    and the output is then exported into JSON using pandas's JSON export
    function. Eventually, we will aim to create a thoroughbred JSON export
    function.
    """

    def __init__(self,
                 orient="records",
                 date_format="iso"):
        """
        Creates a JSON handler. Accepts an encoding orientation and a datetime
        format setting.

        Possible encoding formats are the same as for the `pandas` `DataFrame`
        `to_csv` method:
        - split: one dict per record with columns and data as separate fields
        - records (default): one array per record, with columns as separate
        dicts
        - index: one dict per record, assigning a dict of column-value pairs
        to the index
        - columns: one dict per column, assigning a dict of index-value pairs
        to the column name
        - values: just the values array

        Possible datetime formats are:
        - iso (default): ISO8601 date format
        - epoch: UNIX epoch milliseconds format

        :param orient: encoding orientation
        :param date_format: datetime format
        """
        self.orient = orient
        self.date_format = date_format
        self.logger = logging.getLogger("pyrst.client")
        self.logger.info("Setting up JSON handler...")
        self.logger.info("JSON Handler options: date format is {date_format}, orientation is {orientation}."
                         .format(date_format=self.date_format,
                                 orientation=self.orient))

    def process(self,
                query_output):
        """
        Query output processor that returns a JSON string representation.

        :param query_output: raw query output
        :return: the representation of the query results as a JSON string
        :rtype: str
        """

        self.logger.debug("Processing to DataFrame.")
        _df = DfHandler().process(query_output=query_output)
        self.logger.debug("Processing to DataFrame complete.")

        self.logger.debug("Exporting to JSON.")

        res = _df.to_json(orient=self.orient,
                          date_format=self.date_format,
                          double_precision=2)

        return json.loads(res)


class CsvHandler(Handler):
    """
    Handler that returns a CSV file, ready to be ingested by Excel etc..

    Currently, this is implemented 'lazily', i.e. the DfHandler is invoked,
    and the output is then exported into JSON using pandas's CSV export
    function. Eventually, we will aim to create a thoroughbred CSV export
    function.
    """

    def __init__(self,
                 sep=',',
                 encoding="utf-8",
                 index=False):
        """
        Creates a CSV handler.

        :param sep: separator
        :type sep: str
        :param encoding: encoding (utf-8 or ascii)
        :type encoding: str
        :param index: whether to include index column in output (default:
        False)
        :type index: bool
        """
        self.sep = sep
        self.encoding = encoding
        self.index = index
        self.logger = logging.getLogger("pyrst.client")
        self.logger.info("Setting up CSV handler...")
        self.logger.info("CSV Handler options: separated by {separator}, encoding: {encoding}"
                         .format(separator=self.sep,
                                 encoding=self.encoding))

    def process(self,
                query_output):
        """
        Query output processor that returns a CSV as string representation.

        :param query_output: raw query output
        :return: representation of the result as a CSV string
        :rtype: str
        """

        self.logger.debug("Processing to DataFrame.")
        _df = DfHandler().process(query_output=query_output)
        self.logger.debug("Processing to DataFrame complete.")

        self.logger.debug("Exporting to CSV.")
        return _df.to_csv(sep=self.sep,
                          encoding=self.encoding,
                          index=self.index)


# coding=utf-8

from distutils.core import setup

__version__ = "0.6.1"

long_descr = open("README").read()

setup(
    name='pyrst',
    version=__version__,
    packages=['pyrst'],
    url='https://github.com/rbonedata/pyrst',
    license='Apache 2.0',
    keywords=['BI', 'business intelligence', 'api', 'birst'],
    author='Chris von Csefalvay',
    author_email='chris@chrisvoncsefalvay.com',
    description='Pyrst is a Python client for Birst\'s API.',
    long_description=long_descr,
    install_requires=['suds', 'pandas', 'PyYAML'],
    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Console",
                 "Intended Audience :: Financial and Insurance Industry",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved :: Apache Software License",
                 "Operating System :: OS Independent",
                 "Topic :: Office/Business :: Financial"],
    entry_points = {
        "console_scripts": ['pyrst = pyrst.pyrst_cli:main']
    }
)



# coding=utf-8

from suds.client import Client

from base64 import b64decode
import yaml
import types
import logging

from pyrst.exceptions import SpaceIDException, MissingCredentialsException
from pyrst.decorators import check_token
from pyrst.handlers import Handler, JsonHandler, DfHandler, CsvHandler

module_logger = logging.getLogger("pyrst.client")
module_logger.setLevel(logging.ERROR)
default_formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
module_handler = logging.StreamHandler()
module_logger.addHandler(module_handler)


class BirstClient(object):
    """
    Basic Birst client object.
    """

    def __init__(self,
                 user=None,
                 password=None,
                 instance="app2102",
                 configfile=None):
        """
        Creates the Birst client object.

        Birst client objects can be created either by directly supplying the
        username and the base64-encoded password or by putting it in a
        separate configuration file. The latter version is much preferred,
        as that way, the password will not be present in your code, not even
        in a generally non-trivially readable form. It's important to ensure
        that other than the Python process using Pyrst, no other processes or
        users have access to the configuration file.


        :param user: username
        :type user: str
        :param password: password, encrypted in base64
        :type password: str
        :param instance: name of the instance (e.g. 'app2102'). Defaults to
        app2102.
        :type instance: str
        :param configfile: relative path to a configuration file.
        :type configfile: str
        """

        self.logger = module_logger
        self.logger.info("Creating Birst connector...")

        if configfile:
            self.logger.info("Using configuration file {configfile}".format(configfile=configfile))
            with open(configfile) as _c:
                _config_dict = yaml.load(_c)

            self.password = _config_dict["password"]
            self.user = _config_dict["username"]
            instance = _config_dict["instance"]

            self.logger.debug("User: {username} on instance {instancename}".
                              format(username=self.user,
                                     instancename=instance))

            if _config_dict["password_is_encrypted"] is True:
                self.password = b64decode(_config_dict["password"])
            else:
                self.password = _config_dict["password"]

        else:
            self.user = user
            self.password = b64decode(password)

        if not self.password or not self.user:
            raise MissingCredentialsException

        self.instance = "https://{instancename}.bws.birst.com/CommandWebService.asmx?wsdl"\
            .format(instancename=instance)
        self.token = None

        self.connector = Client(self.instance, location=self.instance)
        self.logger.debug("Connector set up successfully.")

    def __repr__(self):
        return "Birst client instance for user {username} at {instance}".format(username=self.user,
                                                                                instance=self.instance)

    ####################
    # LOGIN AND LOGOUT #
    ####################
    #
    # The login API exposes two (fairly self-explanatory) methods:
    # - login
    # - logout
    #
    # There is no manual token handling in Pyrst - upon login, your token will
    # be appended to the instance as an instance variable.

    # login

    def login(self):
        """
        Logs the user in to the instance specified in the basic settings of the
        class with the password and username specified.
        If successful, the token returned will be appended to the class instance
        as an instance variable.

        :return: token
        :rtype: str
        """

        try:
            self.logger.debug("Obtaining token now...")
            self.token = self.connector.service.Login(self.user, self.password)
            self.logger.info("You have been successfully logged in, {username}.\n"
                             "Your token is: {token}".format(username=self.user,
                                                             token=self.token))
            return self.token
        except Exception as e:
            return e

    # logout

    @check_token
    def logout(self):
        """
        Logs the user out and deletes the token saved in the instance.
        """
        try:
            self.logger.info("Logging out {user}...".format(user=self.user))
            self.connector.service.Logout(self.token)
            self.token = None
            self.logger.warn("You have been logged out.")
        except Exception as e:
            return e

    ##################
    # LISTING SPACES #
    ##################
    #
    # Listing spaces the user has access to.

    # listspaces

    @check_token
    def listspaces(self):
        """
        Lists spaces.

        :return: array of dicts, each representing a space.
        :type: list of dict of (str, str, str)
        """
        self.logger.debug("Listing spaces available to user %s..." % self.user)
        p = self.connector.service.listSpaces(self.token).UserSpace

        result = [{"name": each["name"],
                   "owner": each["owner"],
                   "id": each["id"]} for each in p]

        self.logger.info("{} spaces found, listing.".format(len(result)))
        return result

    ############
    # QUERYING #
    ############
    #
    # The query interface exposes two methods:
    # - executequery: simple querying
    # - retrieve: keep querying as long as there are results.
    #
    # Unlike in Birst's XML API, in Pyrst, the space comes _before_ the query -
    # this is much more intuitive since the query is subordinate to the space
    # rather than the other way around.

    # executequery

    @check_token
    def executequery(self,
                     space,
                     query,
                     handler=None):
        """
        Retrieves the first 1,000 results for the query.

        :param space: SpaceID of the space (incl. hyphens, 36 chars)
        :type space: str
        :param query: Birst BQL query
        :type query: str
        :param handler: output handler class or output handler class instance
        :type handler: Handler
        :return: query result as processed by the handler
        """
        if len(space) != 36:
            raise SpaceIDException
        else:
            self.logger.debug("Executing query.")
            self.logger.debug("Query:\n{}".format(query))
            self.logger.debug("Space: {}".format(space))
            self.logger.debug("Handled by {handler_class}."
                              .format(handler_class=handler.__name__ if handler else "raw output"))

            result = self.connector.service.executeQueryInSpace(self.token,
                                                                query,
                                                                space)

        if handler:
            self.logger.debug("Submitting rows to handler {handler_class_name}."
                              .format(handler.__class__.__name__))
            _handler = handler()
            return _handler.process(result)
        else:
            return result

    # retrieve

    @check_token
    def retrieve(self,
                 space,
                 query,
                 handler=None):
        """
        Retrieves the entire dataset for the query, repeating the `queryMore`
        command as long as there are results. Please be aware that for large
        queries, *this may take some time*.

        :param space: SpaceID of the space (incl. hyphens, 36 chars)
        :type space: str
        :param query: Birst BQL query
        :type query: str
        :param handler: output handler class or output handler class instance
        :type handler: Handler
        :return: query result as processed by the query handler.
        """

        if len(space) != 36:
            raise SpaceIDException
        else:
            self.logger.debug("Executing query.")
            self.logger.debug("Query:\n{}".format(query))
            self.logger.debug("Space: {}".format(space))
            self.logger.debug("Handled by {handler_class}."
                              .format(handler_class=handler.__name__ if handler else "raw output"))

            result = self.connector.service.executeQueryInSpace(self.token,
                                                                query,
                                                                space)

            _result_struct = {"columnNames": result.columnNames[0],
                              "rows": result.rows[0],
                              "dataTypes": result.dataTypes[0],
                              "hasMoreRows": result.hasMoreRows}

            while _result_struct['hasMoreRows']:
                _result_struct["queryToken"] = result.queryToken
                _more_query = self.connector.service.queryMore(self.token,
                                                               _result_struct['queryToken'])
                _result_struct["rows"] += _more_query["rows"][0]
                _result_struct["hasMoreRows"] = _more_query["hasMoreRows"]

        if handler:
            self.logger.debug("Submitting rows to handler {}.".format(handler))
            if isinstance(handler, types.TypeType):
                _handler = handler()
                return _handler.process(_result_struct)
            else:
                return handler.process(_result_struct)
        else:
            return _result_struct
